#ifndef HEURISTIC_FUNCTS
#define HEURISTIC_FUNCTS

#include "debug.h"
#include "bd_string.h"

typedef int (*HeurFunc)();	//Data type for a heuristic function; retuns the board's value

#define HEURMAX 2147483646

extern int searchOrder[7];

/*
   Yes, I used a bunch of free-floating functions and you can call functions that you
   really shouldn't mess with, but I'm too lazy to make black-box objects to
   prevent a rogue coder from calling things like mooreHeurEval and bestColForPlayer.
   So sue me. Sometimes the nasty C way is more appealing than the artful C++ way.
*/

HeurFunc promptForHeuristic();	//Asks the user to select a heuristic and then returns what the user entered
String getHeuristicName(HeurFunc heur);	//returns the name of the passed heuristic

int guessBestCol(Square color, int depth, HeurFunc heurFunc);	//Guesses the best move for passed player with passed heuristic and search depth; returns -1 if the game has ended or you passed sqNull or sqEmpty like a dumb-ass

inline int getMiniMaxValue(Square color, int depth, HeurFunc heurFunc, int bestMin, int bestMax)
//Gets the best score for the current state after depth turns if each player aims for the best leaf of the state tree
{
	int currValue;
	int iter;
	int iterCol;
	int iterInc;
	
	Square winner;
	
	Board *board;
	
	board=Board::getInstance();
	
	winner=board->victory();
	
	if(depth==0 || winner!=sqNull)
	//Return value if search depth is reached or if game is at an end in this state
	{
		if(winner==sqWhite) return HEURMAX;
		if(winner==sqBlack) return -HEURMAX;
		if(winner==sqEmpty) return 0;
		
		return heurFunc();
	}
	
	if(board->isSymmetrical()) iterInc=1;
	else iterInc=1;
	
	for(iter=0, iterCol=searchOrder[0]; iter<COLS; iter++, iterCol=searchOrder[iter])
	{
		if(board->move(iterCol, color))
		{
			currValue=getMiniMaxValue((color==sqWhite)?sqBlack:sqWhite, depth-1, heurFunc, bestMin, bestMax);
			
			if(color==sqBlack && currValue<=bestMax)
			{
				bestMax=currValue;
				
				if(bestMax<=bestMin)
				{
					board->undo();
					
					return bestMax;
				}
			}
			else if(color==sqWhite && currValue>=bestMin)
			{
				bestMin=currValue;
				if(bestMax<=bestMin)
				{
					board->undo();
					
					return bestMin;
				}
			}
			
			board->undo();
		}
	}
	
	return (color==sqBlack)?bestMax:bestMin;
}

inline int mooreHeurEval(int black, int white)
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1;
		if(black==2) return -10;
		return -50;
	}
	
	if(white==1) return 1;
	if(white==2) return 10;
	return 50;
}

inline int mooreHeur()
//Dr. Moore's proposed heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=mooreHeurEval(black, white);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=mooreHeurEval(black, white);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=mooreHeurEval(black, white);
	}
	
	return heur;
}

inline int testHeur1Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur1()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -10 << emptyPlayable;
		return -50 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 10 << emptyPlayable;
	return 50 << emptyPlayable;
}

inline int testHeur1()
//My first test heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur1Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur1Eval(black, white, 0);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur1Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur2Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur2()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -10 << emptyPlayable;
		return -500 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 10 << emptyPlayable;
	return 500 << emptyPlayable;
}

inline int testHeur2()
//My second heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur2Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur2Eval(black, white, 0);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur2Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur3Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur3()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -20 << emptyPlayable;
		return -1000 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 20 << emptyPlayable;
	return 1000 << emptyPlayable;
}

inline int testHeur3()
//My third heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur3Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur3Eval(black, white, 0);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur3Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur4Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur4()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -50 << emptyPlayable;
		return -10000 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 50 << emptyPlayable;
	return 10000 << emptyPlayable;
}

inline int testHeur4()
//My fourth heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur4Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur4Eval(black, white, 1);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur4Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur5Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur5()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -2 << emptyPlayable;
		return -50000 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 2 << emptyPlayable;
	return 50000 << emptyPlayable;
}

inline int testHeur5()
//My fifth heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur5Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur5Eval(black, white, 0);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur5Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur6Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur6()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -2 << emptyPlayable;
		return -50000 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 2 << emptyPlayable;
	return 50000 << emptyPlayable;
}

inline int testHeur6()
//My sixth heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur6Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur6Eval(black, white, 1);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur6Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

inline int testHeur7Eval(int black, int white, int emptyPlayable)
//Used internally by testHeur7()
{
	if((!black && !white) || (black && white)) return 0;
	
	if(black)
	{
		if(black==1) return -1 << emptyPlayable;
		if(black==2) return -10 << emptyPlayable;
		return -50 << emptyPlayable;
	}
	
	if(white==1) return 1 << emptyPlayable;
	if(white==2) return 10 << emptyPlayable;
	return 50 << emptyPlayable;
}

inline int testHeur7()
//My seventh heuristic
{
	int iter1;
	int iter2;
	
	int white;
	int black;
	int emptyPlayable;
	
	int heur;
	
	Board *board;
	
	board=Board::getInstance();
	
	heur=0;
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->rows[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerRows[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur7Eval(black, white, emptyPlayable);
	}
	
	for(iter1=0; iter1<21; iter1++)
	{
		white=0;
		black=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->cols[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
			}
		}
		
		heur+=testHeur7Eval(black, white, 1);
	}
	
	for(iter1=0; iter1<24; iter1++)
	{
		white=0;
		black=0;
		emptyPlayable=0;
		
		for(iter2=0; iter2<4; iter2++)
		{
			switch(*board->diags[iter1][iter2])
			{
				case sqWhite:
				{
					white++;
					break;
				}
				case sqBlack:
				{
					black++;
					break;
				}
				default:
				{
					if(*board->lowerDiags[iter1][iter2]!=sqEmpty) emptyPlayable++;
					break;
				}
			}
		}
		
		heur+=testHeur7Eval(black, white, emptyPlayable);
	}
	
	return heur;
}

#endif
